from __future__ import annotations

import logging
from collections.abc import Mapping

import numpy as np
import pygeomhpges
from numpy.typing import ArrayLike, NDArray
from scipy.stats import rv_continuous

from revertex import core, utils

log = logging.getLogger(__name__)


def sample_hpge_surface(
    n_tot: int,
    seed: int | None = None,
    *,
    hpges: dict[str, pygeomhpges.HPGe] | pygeomhpges.HPGe,
    positions: dict[str, ArrayLike] | ArrayLike,
    surface_type: str | None = None,
) -> NDArray:
    """Generate events on many HPGe's weighting by the surface area.

    Parameters
    ----------
    n_tot
        total number of events to generate
    hpges
        List of :class:`pygeomhpges.HPGe` objects.
    positions
        List of the origin position of each HPGe.
    surface_type
        Which surface to generate events on either `nplus`, `pplus`, `passive` or None (generate on all surfaces).
    seed
        seed for random number generator.

    Returns
    -------
    Array of global coordinates.
    """

    rng = np.random.default_rng(seed=seed)

    out = np.full((n_tot, 3), np.nan)

    # loop over n_det maybe could be faster
    if isinstance(hpges, Mapping):
        # index of the surfaces per detector
        p_det = utils.get_surface_weights(hpges, surface_type=surface_type)
        det_index = rng.choice(np.arange(len(hpges)), size=n_tot, p=p_det)

        for idx, (name, hpge) in enumerate(hpges.items()):
            n = np.sum(det_index == idx)

            out[det_index == idx] = (
                _sample_hpge_surface_impl(n, hpge, surface_type=surface_type, seed=seed)
                + positions[name]
            )
    else:
        out = (
            _sample_hpge_surface_impl(
                n_tot, hpges, surface_type=surface_type, seed=seed
            )
            + positions
        )

    return out


def _sample_hpge_surface_impl(
    n: int,
    hpge: pygeomhpges.HPGe,
    surface_type: str | None,
    depth: rv_continuous | None = None,
    seed: int | None = None,
) -> NDArray:
    """Generate events on the surface of a single HPGe.

    Parameters
    ----------
    n
        number of vertexs to generate.
    hpge
        pygeomhpges object describing the detector geometry.
    surface_type
        Which surface to generate events on either `nplus`, `pplus`, `passive` or None (generate on all surfaces).
    depth
        scipy `rv_continuous` object describing the depth profile, if None events are generated directly on the surface.
    seed
        seed for random number generator.

    Returns
    -------
    Array with shape `(n,3)` describing the local `(x,y,z)` positions for every vertex
    """
    rng = (
        np.random.default_rng(seed=seed)
        if seed is not None
        else np.random.default_rng()
    )

    surface_indices = utils.get_surface_indices(hpge, surface_type)

    # surface areas
    areas = hpge.surface_area(surface_indices).magnitude

    # get the sides
    sides = rng.choice(surface_indices, size=n, p=areas / np.sum(areas))
    # get thhe detector geometry
    r, z = hpge.get_profile()
    s1, s2 = pygeomhpges.utils.get_line_segments(r, z)

    # compute random coordinates
    r1 = s1[sides][:, 0]
    r2 = s2[sides][:, 0]

    frac = core.sample_proportional_radius(r1, r2, size=(len(sides)))

    rz_coords = s1[sides] + (s2[sides] - s1[sides]) * frac[:, np.newaxis]

    phi = rng.uniform(low=0, high=2 * np.pi, size=(len(sides)))

    # convert to random x,y
    x = rz_coords[:, 0] * np.cos(phi)
    y = rz_coords[:, 0] * np.sin(phi)

    if depth is not None:
        msg = "depth profile is not yet implemented "
        raise NotImplementedError(msg)

    return np.vstack([x, y, rz_coords[:, 1]]).T



from __future__ import annotations

import numpy as np
import pygeomhpges
from pyg4ometry import geant4

from revertex.generators.shell import (
    _sample_hpge_shell_impl,
    sample_hpge_shell,
)


def test_shell_gen(test_data_configs):
    hpge = pygeomhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=None)

    coords = _sample_hpge_shell_impl(100, hpge, surface_type=None, distance=1)

    assert np.shape(coords) == (100, 3)


def test_many_shell_gen(test_data_configs):
    reg = geant4.Registry()
    hpge_IC = pygeomhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=reg)
    hpge_BG = pygeomhpges.make_hpge(test_data_configs + "/B99000A.yaml", registry=reg)
    hpge_SC = pygeomhpges.make_hpge(test_data_configs + "/C99000A.yaml", registry=reg)

    coords = sample_hpge_shell(
        1000,
        seed=None,
        distance=2,
        hpges={"V99000A": hpge_IC, "B99000A": hpge_BG, "C99000A": hpge_SC},
        positions={"V99000A": [0, 0, 0], "B99000A": [0, 0, 0], "C99000A": [0, 0, 0]},
    )

    assert np.shape(coords) == (1000, 3)

    # should also work for one hpge

    coords = sample_hpge_shell(
        1000, seed=None, distance=2, hpges=hpge_IC, positions=[0, 0, 0]
    )

    assert np.shape(coords) == (1000, 3)



from __future__ import annotations

import logging

import awkward as ak
import hist
import numpy as np
from lgdo import lh5
from numpy.typing import ArrayLike
from scipy.spatial.transform import Rotation as rot

from revertex import utils
from revertex.core import _get_chunks, convert_output, sample_histogram

log = logging.getLogger(__name__)


def save_beta_spectrum(
    n_gen: int, in_file: str, out_file: str, seed: int | None = None, eunit: str = "keV"
) -> None:
    """Save positions generated by the function to a file.

    Parameters
    ----------
    func
        function which generates the positions, must return an `ak.Array` of positions.
    n_gen
        number of positions to generate
    in_file
        path to the CSV theory file.
    out_file
        path to the output file.
    seed
        random seed, for each chunk will be multiplied by 7.
    lunit
        The length unit returned by the function.
    **kwargs
        keyword arguments to be passed to the function.
    """
    # read input
    energies, phase_space = utils.read_input_beta_csv(in_file, delimiter=",")

    msg = f"Read beta spectrum from {in_file} E = {energies}, phase space = {phase_space} "
    log.info(msg)

    chunks = _get_chunks(n_gen, 1000_000)

    for idx, chunk in enumerate(chunks):
        # generate kinematics
        kin_ak = generate_beta_spectrum(
            chunk, energies=energies, phase_space=phase_space, seed=seed, eunit=eunit
        )
        msg = f"Generated beta kinematics {kin_ak}"
        log.debug(msg)

        # update the seed
        seed = seed * 7 if seed is not None else None

        # convert
        kin_lh5 = convert_output(kin_ak, mode="kin", eunit=eunit)

        # write
        mode = "of" if idx == 0 else "append"
        lh5.write(kin_lh5, "vtx/kin", out_file, wo_mode=mode)


def generate_beta_spectrum(
    size: int,
    *,
    energies: ArrayLike,
    phase_space: ArrayLike,
    seed: int | None = None,
    eunit: str = "keV",
) -> ak.Array:
    """Generate samples from a beta spectrum defined by a list of energies and phase space
    values.

    This function interprets the energies and phase_space as a histogram and samples
    energies from this. These are then converted into momenta. The energies should be
    the left edge of the bin of a histogram

    Parameters
    ----------
    energies
        the energy values
    spectrum
        the phase space values
    size
        number of events to generate
    seed
        random seed.
    eunit
        the unit for energy in the input file, default keV.

    Returns
    -------
    An awkward array with the sampled kinematics, in keV.
    """
    # convert energies to bin edges
    if eunit == "MeV":
        factor = 1 / 1000.0
    elif eunit == "keV":
        factor = 1
    elif eunit == "eV":
        factor = 1000
    else:
        msg = f"Only eunits keV, MeV or eV are supported not {eunit}"
        raise ValueError(msg)

    histo = hist.Hist(hist.axis.Variable(energies * factor))

    for b in range(histo.size - 2):
        histo[b] = phase_space[b]

    energy_samples = sample_histogram(histo, size, seed=seed)
    matrix = np.vstack(
        [[energy_samples, np.zeros_like(energy_samples), np.zeros_like(energy_samples)]]
    )

    mass = 511.0  # keV
    momentum = np.sqrt(energy_samples**2 + 2 * mass * energy_samples)

    rng = np.random.default_rng(seed)  # Set seed
    rand_rot = rot.random(random_state=rng)  # Random rotation

    matrix = np.vstack([[momentum, np.zeros_like(momentum), np.zeros_like(momentum)]])
    momenta = rand_rot.as_matrix() @ matrix

    return ak.Array(
        {
            "px": momenta[0, :],
            "py": momenta[1, :],
            "pz": momenta[2, :],
            "ekin": energy_samples,
            "g4_pid": np.full_like(energy_samples, 11),
        }
    )



from __future__ import annotations

import logging
from collections.abc import Mapping

import numpy as np
import pygeomhpges
from numpy.typing import ArrayLike, NDArray

from revertex import core, utils

log = logging.getLogger(__name__)


def sample_hpge_shell(
    n_tot: int,
    *,
    seed: int | None = None,
    hpges: dict[str, pygeomhpges.HPGe] | pygeomhpges.HPGe,
    positions: dict[str, ArrayLike] | ArrayLike,
    distance: float,
    surface_type: str | None = None,
) -> NDArray:
    """Generate events on many HPGe's shells weighting by the surface area.

    Parameters
    ----------
    n_tot
        total number of events to generate
    seed
        random seed for the RNG.
    hpges
        List of :class:`pygeomhpges.HPGe` objects.
    positions
        List of the origin position of each HPGe.
    surface_type
        Which surface to generate events on either `nplus`, `pplus`, `passive` or None (generate on all surfaces).

    Returns
    -------
    Array of global coordinates.
    """

    rng = np.random.default_rng(seed=seed)

    out = np.full((n_tot, 3), np.nan)

    if isinstance(hpges, Mapping):
        p_det = utils.get_surface_weights(hpges, surface_type=surface_type)
        det_index = rng.choice(np.arange(len(hpges)), size=n_tot, p=p_det)

        for idx, (name, hpge) in enumerate(hpges.items()):
            n = np.sum(det_index == idx)

            out[det_index == idx] = (
                _sample_hpge_shell_impl(
                    n, hpge, distance=distance, surface_type=surface_type, seed=seed
                )
                + positions[name]
            )

    else:
        out = (
            _sample_hpge_shell_impl(
                n_tot, hpges, distance=distance, surface_type=surface_type, seed=seed
            )
            + positions
        )
    return out


def _sample_hpge_shell_impl(
    size: int,
    hpge: pygeomhpges.HPGe,
    surface_type: str | None,
    distance: float,
    seed: int | None = None,
) -> NDArray:
    """Generate events on a shell around a single HPGe. This uses rejection sampling.

    Parameters
    ----------
    size
        number of vertexs to generate.
    hpge
        pygeomhpges object describing the detector geometry.
    surface_type
        Which surface to generate events on either `nplus`, `pplus`, `passive` or None (generate on all surfaces).
    distance
        Size of the hpge shell to generate in.
    seed
        seed for random number generator.

    Returns
    -------
    Array with shape `(n,3)` describing the local `(x,y,z)` positions for every vertex
    """

    # get the surface indices (which sides to use)
    surface_indices = utils.get_surface_indices(hpge, surface_type)

    # the bounding box should be in x +/- (radius+distance)
    # and in y -distance to height + distance

    r, z = hpge.get_profile()

    height = max(z)
    radius = max(r)

    output = None

    # sampling efficiency is not necessarily high but hopefully this is not a big limitation
    seed_tmp = seed

    while output is None or (len(output) < size):
        # adjust seed
        seed_tmp = seed_tmp * 7 if seed_tmp is not None else seed

        # get some proposed points
        proposals = core.sample_cylinder(
            r_range=(0, radius + distance),
            z_range=(-distance, height + distance),
            size=size * 5,
            seed=seed_tmp,
        )

        distances = hpge.distance_to_surface(proposals, surface_indices, signed=True)

        # should be negative (outside) and > -distance
        is_good = (distances < 0) & (abs(distances) < distance)
        sel = proposals[is_good]

        # extend
        output = np.vstack((output, sel)) if output is not None else sel

    # now cut to the right size
    return output[:size]



from __future__ import annotations

import logging
from collections.abc import Mapping

import numpy as np
import pygeomhpges
from numpy.typing import ArrayLike, NDArray

from revertex import core, utils

log = logging.getLogger(__name__)


def sample_hpge_borehole(
    n_tot: int,
    *,
    seed: int | None = None,
    hpges: dict[str, pygeomhpges.HPGe] | pygeomhpges.HPGe,
    positions: dict[str, ArrayLike] | ArrayLike,
) -> NDArray:
    """Generate events on many HPGe boreholes weighting by the volume.

    Parameters
    ----------
    n_tot
        total number of events to generate
    seed
        random seed for the RNG.
    hpges
        List of :class:`pygeomhpges.HPGe` objects.
    positions
        List of the origin position of each HPGe.

    Returns
    -------
    Array of global coordinates.
    """
    rng = np.random.default_rng(seed=seed)

    out = np.full((n_tot, 3), np.nan)

    # loop over n_det maybe could be faster
    if isinstance(hpges, Mapping):
        weights = utils.get_borehole_weights(hpges)

        det_index = rng.choice(np.arange(len(hpges)), size=n_tot, p=weights)

        for idx, (name, hpge) in enumerate(hpges.items()):
            n = np.sum(det_index == idx)

            out[det_index == idx] = (
                _sample_hpge_borehole_impl(n, hpge, seed=seed) + positions[name]
            )
    else:
        out = _sample_hpge_borehole_impl(n_tot, hpges, seed=seed) + positions

    return out


def _sample_hpge_borehole_impl(
    size: int,
    hpge: pygeomhpges.HPGe,
    seed: int | None = None,
) -> NDArray:
    """Generate events on the surface of a single HPGe.

    Parameters
    ----------
    n
        number of vertexs to generate.
    hpge
        pygeomhpges object describing the detector geometry.
    surface_type
        Which surface to generate events on either `nplus`, `pplus`, `passive` or None (generate on all surfaces).
    seed
        seed for random number generator.

    Returns
    -------
    Array with shape `(n,3)` describing the local `(x,y,z)` positions for every vertex
    """
    r, z = hpge.get_profile()

    height = max(z)
    radius = max(r)

    output = None

    # sampling efficiency is not necessarily high but hopefully this is not a big limitation
    seed_tmp = seed

    while output is None or (len(output) < size):
        # adjust seed
        seed_tmp = seed_tmp * 7 if seed_tmp is not None else seed

        # get some proposed points
        proposals = core.sample_cylinder(
            r_range=(0, radius),
            z_range=(0, height),
            size=size,
            seed=seed_tmp,
        )

        is_good = hpge.is_inside_borehole(proposals)

        sel = proposals[is_good]

        # extend
        output = np.vstack((output, sel)) if output is not None else sel

    # now cut to the right size
    return output[:size]



from __future__ import annotations

import numpy as np
import pygeomhpges

from revertex.generators.borehole import (
    _sample_hpge_borehole_impl,
    sample_hpge_borehole,
)


def test_borehole_gen(test_data_configs):
    hpge = pygeomhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=None)

    coords = _sample_hpge_borehole_impl(100, hpge)

    assert np.shape(coords) == (100, 3)

    assert np.shape(
        sample_hpge_borehole(
            1000, seed=None, hpges={"IC": hpge}, positions={"IC": [0, 0, 0]}
        )
    ) == (1000, 3)

    # should also work for one hpge

    coords = sample_hpge_borehole(1000, seed=None, hpges=hpge, positions=[0, 0, 0])

    assert np.shape(coords) == (1000, 3)



from __future__ import annotations

import logging
import re

import colorlog
import numpy as np
import pyg4ometry.geant4 as pg4
import pygeomhpges
import pygeomtools

from revertex import core

log = logging.getLogger(__name__)


def expand_regex(inputs: list, patterns: list) -> list:
    """Get a list of detectors from regex

    This matches any wildcars with * or ? in the patterns.

    Parameters
    ----------
    inputs
        list of input strings to find matches in.
    patterns
        list of patterns to search for.
    """
    regex_patterns = [
        re.compile(
            "^" + p.replace(".", r"\.").replace("*", ".*").replace("?", ".") + "$"
        )
        for p in patterns
    ]
    return [v for v in inputs if any(r.fullmatch(v) for r in regex_patterns)]


def read_input_beta_csv(path: str, **kwargs) -> tuple[np.ndarray, np.ndarray]:
    """Reads a CSV file into numpy arrays.

    The file should have the following format:

        energy_1, phase_space_1
        energy_2, phase_space_2
        energy_3, phase_space_3

    Parameters
    ----------
    path
        filepath to the csv file.
    kwargs
        keyword arguments to pass to `np.genfromtxt`
    """
    return np.genfromtxt(path, **kwargs).T[0], np.genfromtxt(path, **kwargs).T[1]


def get_hpges(
    reg: pg4.geant4.registry, detectors: str | list[str]
) -> tuple[dict, dict]:
    """Extract the objects for each HPGe detector in `reg` and in the list of `detectors`"""

    phy_vol_dict = reg.physicalVolumeDict
    det_list = expand_regex(list(phy_vol_dict.keys()), list(detectors))

    hpges = {
        name: pygeomhpges.make_hpge(
            pygeomtools.get_sensvol_metadata(reg, name), registry=None
        )
        for name in det_list
    }

    pos = {name: phy_vol_dict[name].position.eval() for name in det_list}

    return hpges, pos


def get_surface_indices(hpge: pygeomhpges.base.HPGe, surface_type: str | None) -> tuple:
    """Get which surface index corresponds to the desired surface type"""

    surf = np.array(hpge.surfaces)
    return (
        np.where(surf == surface_type)[0]
        if (surface_type is not None)
        else np.arange(len(hpge.surfaces))
    )


def get_surface_weights(hpges: dict, surface_type: str | None) -> list:
    """Get a weighting for each hpge in the `hpges` based on surface area
    for a given `surface_type`
    """

    # index of the surfaces per detector
    surf_ids_tot = [
        np.array(hpge.surfaces) == surface_type
        if surface_type is not None
        else np.arange(len(hpge.surfaces))
        for name, hpge in hpges.items()
    ]

    # total surface area per detector
    surf_tot = [
        np.sum(hpge.surface_area(surf_ids).magnitude)
        for (name, hpge), surf_ids in zip(hpges.items(), surf_ids_tot, strict=True)
    ]

    return surf_tot / np.sum(surf_tot)


def get_borehole_volume(hpge: pygeomhpges.HPGe, size=1000000):
    """Estimate the borehole volume (with MC)"""

    r, z = hpge.get_profile()
    height = max(z)
    radius = max(r)

    points = core.sample_cylinder(
        r_range=(0, radius), z_range=(0, height), seed=None, size=size
    )
    vol = np.pi * radius**2 * height

    is_good = len(points[hpge.is_inside_borehole(points)])

    return (is_good / size) * vol


def get_borehole_weights(hpges: dict) -> list:
    """Get a weighting for each hpge in the `hpges` based on borehole volume"""

    vol_tot = [get_borehole_volume(hpge, size=int(1e6)) for _, hpge in hpges.items()]

    return vol_tot / np.sum(vol_tot)


def setup_log(level: int | None = None) -> None:
    """Setup a colored logger for this package.

    Parameters
    ----------
    level
        initial log level, or ``None`` to use the default.
    """
    fmt = "%(log_color)s%(name)s [%(levelname)s]"
    fmt += " %(message)s"

    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter(fmt))

    logger = logging.getLogger("revertex")
    # logger.addHandler(handler)

    if level is not None:
        logger.setLevel(level)



from __future__ import annotations

import logging
from collections.abc import Callable

import awkward as ak
import hist
import numpy as np
from lgdo import lh5
from lgdo.types import Array, Table
from numpy.typing import ArrayLike

log = logging.getLogger(__name__)


def _get_chunks(n: int, m: int) -> np.ndarray:
    return (
        np.full(n // m, m, dtype=int)
        if n % m == 0
        else np.append(np.full(n // m, m, dtype=int), n % m)
    )


def sample_cylinder(
    r_range: float,
    z_range: tuple,
    size: int,
    seed: int | None,
    phi_range: tuple = (0, 2 * np.pi),
):
    """Generate points in a cylinder, returns the points as a 2D array

    Parameters
    ----------
    r_range
        The range of `r` to sample.
    z_range
        The range of `z` to sample.
    phi
        The range of angles to sample.
    size
        The number of points to generate.
    seed
        The random seed for the rng.
    """

    rng = np.random.default_rng(seed=seed)

    r2 = rng.uniform(low=r_range[0] ** 2, high=r_range[1] ** 2, size=size)
    r = np.sqrt(r2)

    z = rng.uniform(low=z_range[0], high=z_range[1], size=size)
    phi = rng.uniform(low=phi_range[0], high=phi_range[1], size=size)

    x = r * np.cos(phi)
    y = r * np.sin(phi)

    return np.column_stack((x, y, z))


def sample_histogram(
    histo: hist.Hist, size: int, *, seed: int | None = None
) -> np.ndarray:
    """Generate samples from a 1D or 2D histogram.

    Based on approximating the histogram as a piecewise uniform,
    probability distribution.

    Parameters
    ----------
    histo
        The histogram to generate samples from.
    size
        The number of samples to generate.
    seed
        Random seed.

    Returns
    -------
    an array of the samples (1D case) of a tuple of x, y samples (2D case)
    """
    # create rng
    rng = np.random.default_rng(seed=seed)

    if not isinstance(histo, hist.Hist):
        msg = f"sample histogram needs hist.Hist object not {type(histo)}"
        raise TypeError(msg)

    ndim = histo.ndim

    if ndim == 1:
        # convert to numpy
        probs, bins = histo.to_numpy()

        # compute the binwidth
        binwidths = np.diff(bins)

        # normalise
        probs = probs / sum(probs)
        bin_idx = rng.choice(np.arange(len(probs)), size=size, p=probs)
        start = bins[bin_idx]

        # the value within the bin
        delta = rng.uniform(low=0, high=1, size=size) * binwidths[bin_idx]
        return start + delta

    if ndim == 2:
        probs, bins_x, bins_y = histo.to_numpy()

        binwidths_x = np.diff(bins_x)
        binwidths_y = np.diff(bins_y)

        # get the flattened bin number
        probs = probs / np.sum(probs)
        bin_idx = rng.choice(np.arange(np.size(probs)), size=size, p=probs.flatten())

        _rows, cols = probs.shape

        # extract unflattened indices
        row = bin_idx // cols
        col = bin_idx % cols

        # get the delta within the bin
        delta_x = rng.uniform(low=0, high=1, size=size) * binwidths_x[row]
        delta_y = rng.uniform(low=0, high=1, size=size) * binwidths_y[col]

        # returned values
        values_x = delta_x + bins_x[row]
        values_y = delta_y + bins_y[col]

        return values_x, values_y

    msg = f"It is only supported to sample from 1D or 2D histograms not {ndim}"
    raise ValueError(msg)


def convert_output(
    arr: ak.Array, *, mode: str = "pos", lunit: str = "mm", eunit: str = "keV"
) -> Table:
    """Converts the vertices to the correct output format.

    This function creates a table of either `kin` or `pos` information.

    Parameters
    ----------
    arr
        The input data to convert
    mode
        The mode either 'pos' or 'kin'
    lunit
        Unit for distances, by default mm.
    eunit
        Unit for energy, by default keV.

    Returns
    -------
    The output table.
    """
    out = Table(size=len(arr))

    if mode == "pos":
        for field in ["xloc", "yloc", "zloc"]:
            out.add_field(field, Array(arr[field].to_numpy(), attrs={"units": lunit}))

    elif mode == "kin":
        for field in ["px", "py", "pz", "ekin"]:
            unit = eunit if field == "ekin" else ""
            out.add_field(field, Array(arr[field].to_numpy(), attrs={"units": unit}))

        out.add_field(
            "g4_pid", Array(arr["g4_pid"].to_numpy().astype(np.int64), dtype=np.int64)
        )
    else:
        msg = f"Only modes pos or kin are supported for converting outputs not {mode}"
        raise ValueError(msg)

    return out


def write_remage_vtx(
    n: int,
    out_file: str,
    seed: int | None,
    generator: Callable,
    lunit: str = "mm",
    **kwargs,
) -> None:
    """Save the vertices generatored by a particular vertex generator function.

    This follows the convention ` defined here <https://remage.readthedocs.io/en/stable/manual/generators.html#simulating-event-vertices-and-kinematics-from-external-files>`_

    Parameters
    ----------
    n
        The number of vertices to generate
    out_file
        The path to the file to save the results.
    seed
        The seed to the random number generator
    generator
        A function generating the vertices (following the revertex specifications)
    kwargs
        The keyword arguments to the function

    """

    chunks = _get_chunks(n, 1000_000)

    for idx, chunk in enumerate(chunks):
        positions = generator(chunk, **kwargs)

        pos_ak = ak.Array(
            {"xloc": positions[:, 0], "yloc": positions[:, 1], "zloc": positions[:, 2]}
        )

        msg = f"Generated vertices {pos_ak}"
        log.debug(msg)

        # update the seed
        seed = seed * 7 if seed is not None else None

        # convert
        pos_lh5 = convert_output(pos_ak, mode="pos", lunit=lunit)

        msg = f"Output {pos_lh5}"
        log.debug(msg)

        # write
        mode = "of" if idx == 0 else "append"
        lh5.write(pos_lh5, "vtx/pos", out_file, wo_mode=mode)


def sample_proportional_radius(
    r0: ArrayLike, r1: ArrayLike, size: int = 10000, seed: int | None = None
):
    r"""Sample from a distribution weighted by the radius. This is used for the surface sampling og shapes.

    Based on sampling from a distribution:

    .. math::

        P(r) \propto r

    restricted to the range min(r0,r1) to max(r0,r1).


    Parameters
    ----------
    r0
        list of first radius, must have the same length as size.
    r1
        list of second, must have the same length as size.
    size
        number of samples.
    seed
        random seed for rng.
    """
    rng = (
        np.random.default_rng(seed=seed)
        if seed is not None
        else np.random.default_rng()
    )
    if len(r0) != size or len(r1) != size:
        msg = (
            f"r0 and r1 must have {size} elements not {len(r0)} (r0) or {len(r1)} (r1)"
        )
        raise ValueError(msg)

    # Ensure r0 and r1 are numpy arrays
    r0, r1 = np.asarray(r0), np.asarray(r1)

    # Get min and max for each pair
    sign = r1 > r0
    a = np.minimum(r0, r1)
    b = np.maximum(r0, r1)

    # Generate uniform samples for each pair
    u = rng.uniform(size=a.shape)  # Same shape as r0 and r1

    # Apply inverse transform sampling element-wise
    result = u
    mask = a != b

    result[mask] = (
        np.sqrt(u[mask] * (b[mask] ** 2 - a[mask] ** 2) + a[mask] ** 2) - a[mask]
    ) / (b[mask] - a[mask])

    result[~sign] = 1 - result[~sign]

    return result
